import numpy as np
from scipy.signal import convolve, gaussian
from scipy.stats import norm


def simulate_speed(time, sampling_frequency, ripple_times, sigma=0.010,
                   ripple_width=0.050):
    ripple_start_end = (ripple_times[:, np.newaxis]
                        + ripple_width * np.array([-1, 1]))

    is_replay = np.zeros_like(time, dtype=bool)

    for start, end in ripple_start_end:
        is_replay[(time >= start) & (time <= end)] = True
    speed = (is_replay.copy() * -19.0) + 20
    bandwidth = sigma * sampling_frequency
    n_time_window_samples = int(bandwidth * 8)
    kernel = gaussian(n_time_window_samples, bandwidth)
    speed = convolve(speed, kernel, mode='same') / kernel.sum()
    return norm.rvs(loc=speed)



import numpy as np
from patsy import dmatrices
from scipy.stats import norm
from scipy.ndimage.filters import gaussian_filter
from statsmodels.api import GLM, families
from statsmodels.tsa.tsatools import lagmat


def estimate_movement_variance(position, lagged_position, speed):

    data = {
        'position': position,
        'lagged_position': lagged_position
    }

    MODEL_FORMULA = 'position ~ lagged_position - 1'
    response, design_matrix = dmatrices(MODEL_FORMULA, data)
    fit = GLM(response, design_matrix, family=families.Gaussian()).fit()

    return np.sqrt(fit.scale)


def estimate_movement_state_transition(place_bins, position, variance):
    """Zero mean random walk with covariance based on movement.

    p(x_{k} | x_{k-1}, I_{k}, I_{k-1})
    """
    state_transition_matrix = norm.pdf(
        place_bins[:, np.newaxis], loc=place_bins[np.newaxis, :],
        scale=variance)
    return _normalize_row_probability(state_transition_matrix)


def fit_movement_state_transition(position, speed, place_bins,
                                  movement_threshold=4.0, speed_up_factor=20):
    movement_variance = estimate_movement_variance(
        position, speed, movement_threshold)
    return np.linalg.matrix_power(estimate_movement_state_transition(
        place_bins, position, movement_variance), speed_up_factor)


def _normalize_row_probability(x):
    '''Ensure the state transition matrix rows sum to 1
    '''
    return x / x.sum(axis=1, keepdims=True)


def empirical_movement_transition_matrix(position, place_bin_edges, speed,
                                         replay_speedup=16,
                                         movement_threshold=4.0):
    '''Estimate the probablity of the next position based on the movement
     data, given the movment is sped up by the
     `sequence_compression_factor`

    position cell firing during a hippocampal replay event is a "sped-up"
    version of position cell firing when the animal is actually moving.
    Here we use the animal's actual movements to constrain which position
    cell is likely to fire next.

    Parameters
    ----------
    position : array_like, shape (n_time,)
        Linearized position of the animal over time
    place_bin_edges : array_like, shape (n_bins,)
    sequence_compression_factor : int, optional
        How much the movement is sped-up during a replay event
    is_movement : array_like, shape (n_time,)
        Boolean indicator for an experimental condition.

    Returns
    -------
    empirical_movement_transition_matrix : array_like,
                                           shape=(n_bin_edges-1,
                                           n_bin_edges-1)

    '''

    is_movement = speed > movement_threshold
    lagged_place = lagmat(position, 1)[is_movement].squeeze()
    position = position[is_movement]

    movement_variance = estimate_movement_variance(
        position, lagged_place, speed)

    movement_bins, _, _ = np.histogram2d(lagged_place, position,
                                         bins=(place_bin_edges,
                                               place_bin_edges),
                                         normed=False)

    smoothed_movement_bins_probability = gaussian_filter(
        _normalize_row_probability(
            _fix_zero_bins(movement_bins)), sigma=movement_variance)
    return np.linalg.matrix_power(
        smoothed_movement_bins_probability, replay_speedup)


def _fix_zero_bins(movement_bins):
    '''If there is no data observed for a column, set everything to 1 so
    that it will have equal probability
    '''
    movement_bins[:, movement_bins.sum(axis=0) == 0] = 1
    return movement_bins



'''Tools for simulating place field spiking'''
import numpy as np
from scipy.stats import norm


def simulate_time(n_samples, sampling_frequency):
    return np.arange(n_samples) / sampling_frequency


def simulate_linear_distance(time, track_height, running_speed=10):
    half_height = (track_height / 2)
    return (half_height * np.sin(2 * np.pi * time / running_speed - np.pi / 2)
            + half_height)


def simulate_linear_distance_with_pauses(time, track_height, running_speed=10,
                                         pause=0.5, sampling_frequency=1):
    linear_distance = simulate_linear_distance(
        time, track_height, running_speed)
    peaks = np.nonzero(linear_distance == track_height)[0]
    n_pause_samples = int(pause * sampling_frequency)
    pause_linear_distance = np.zeros(
        (time.size + n_pause_samples * peaks.size,))
    pause_ind = (peaks[:, np.newaxis] + np.arange(n_pause_samples))
    pause_ind += np.arange(peaks.size)[:, np.newaxis] * n_pause_samples

    pause_linear_distance[pause_ind.ravel()] = track_height
    pause_linear_distance[pause_linear_distance == 0] = linear_distance
    return pause_linear_distance[:time.size]


def get_trajectory_direction(linear_distance):
    is_inbound = np.insert(np.diff(linear_distance) < 0, 0, False)
    return np.where(is_inbound, 'Inbound', 'Outbound'), is_inbound


def simulate_poisson_spikes(rate, sampling_frequency):
    return 1.0 * (np.random.poisson(rate / sampling_frequency) > 0)


def create_place_field(
    place_field_mean, linear_distance, sampling_frequency, is_condition=None,
        place_field_std_deviation=12.5, max_firing_rate=20,
        baseline_firing_rate=0.1):
    if is_condition is None:
        is_condition = np.ones_like(linear_distance, dtype=bool)
    field_firing_rate = norm(
        place_field_mean, place_field_std_deviation).pdf(linear_distance)
    field_firing_rate /= np.nanmax(field_firing_rate)
    field_firing_rate[~is_condition] = 0
    return baseline_firing_rate + max_firing_rate * field_firing_rate



"""Fitting and predicting the likelihood of replay events based on place field
spiking patterns.
"""

from functools import partial
from logging import getLogger

import numpy as np
import pandas as pd
from patsy import build_design_matrices, dmatrix
from statsmodels.api import families

import dask.array as da
from regularized_glm import penalized_IRLS

from .core import combined_likelihood

try:
    from tqdm import tqdm
except ImportError:
    def tqdm(*args, **kwargs):
        if args:
            return args[0]
        return kwargs.get('iterable', None)


logger = getLogger(__name__)


def fit_glm_model(spikes, design_matrix, penalty=1E1):
    """Fits the Poisson model to the spikes from a neuron.

    Parameters
    ----------
    spikes : array_like
    design_matrix : array_like or pandas DataFrame
    ind : int
    penalty : float, optional

    Returns
    -------
    fitted_model : statsmodel results

    """
    penalty = np.ones((design_matrix.shape[1],)) * penalty
    penalty[0] = 0.0
    results = penalized_IRLS(
        np.array(design_matrix), np.array(spikes),
        family=families.Poisson(), penalty=penalty)
    return np.squeeze(results.coefficients)


def create_predict_design_matrix(position, design_matrix):
    is_nan = np.isnan(position)
    position[is_nan] = 0
    predictors = {'position': position}
    design_matrix = build_design_matrices(
        [design_matrix.design_info], predictors)[0]
    design_matrix[is_nan] = np.nan
    return design_matrix


def get_conditional_intensity(coefficients, design_matrix):
    """Predict the model's response given a design matrix and the model
    parameters.

    Parameters
    ----------
    coefficients : ndarray, shape (n_coefficients, n_neurons)
    design_matrix : ndarray, shape (n_obs, n_coefficients)

    Returns
    -------
    conditional_intensity : ndarray, shape (n_coefficients, n_neurons)

    """
    return np.exp(design_matrix @ coefficients)


def atleast_kd(array, k):
    """
    https://stackoverflow.com/questions/42516569/numpy-add-variable-number-of-dimensions-to-an-array
    """
    new_shape = array.shape + (1,) * (k - array.ndim)
    return array.reshape(new_shape)


@combined_likelihood
def poisson_log_likelihood(is_spike, conditional_intensity=None,
                           time_bin_size=1):
    """Probability of parameters given spiking at a particular time.

    Parameters
    ----------
    is_spike : array_like with values in {0, 1}, shape (n_signals,)
        Indicator of spike or no spike at current time.
    conditional_intensity : array_like, shape (n_signals, n_states,
                                               n_place_bins)
        Instantaneous probability of observing a spike
    time_bin_size : float, optional

    Returns
    -------
    poisson_log_likelihood : array_like, shape (n_signals, n_states,
                                                n_place_bins)

    """
    probability_no_spike = -conditional_intensity * time_bin_size
    is_spike = atleast_kd(is_spike, conditional_intensity.ndim)
    return (np.log(conditional_intensity) * is_spike +
            probability_no_spike) + np.spacing(1)


def spiking_likelihood_ratio(
        is_spike, position, design_matrix, place_field_coefficients,
        place_conditional_intensity, time_bin_size=1, chunks=1E3):
    """Computes the likelihood ratio between replay and not replay events.

    Parameters
    ----------
    is_spike : ndarray, shape (n_time, n_neurons)
    position : ndarray, shape (n_time,)
    design_matrix : ndarray, shape (n_time, n_coefficients)
    place_field_coefficients : ndarray, shape (n_coefficients, n_neurons)
    place_conditional_intensity : ndarray, shape (1, n_place_bins, n_neurons)
    time_bin_size : float, optional

    Returns
    -------
    spiking_likelihood_ratio : ndarray, shape (n_place_bins,)

    """
    no_replay_design_matrix = create_predict_design_matrix(
        position, design_matrix)
    no_replay_conditional_intensity = get_conditional_intensity(
        place_field_coefficients, no_replay_design_matrix)
    no_replay_log_likelihood = da.from_array(poisson_log_likelihood(
        is_spike, no_replay_conditional_intensity,
        time_bin_size)[:, np.newaxis], chunks)
    is_spike = da.from_array(is_spike[:, np.newaxis, :], chunks)
    place_conditional_intensity = da.from_array(
        place_conditional_intensity[np.newaxis, ...], chunks)
    replay_log_likelihood = poisson_log_likelihood(
        is_spike, place_conditional_intensity, time_bin_size)
    return np.exp(replay_log_likelihood - no_replay_log_likelihood).compute()


def fit_spiking_likelihood_ratio(position, spikes, is_replay,
                                 place_bin_centers, penalty=1E1,
                                 time_bin_size=1, knot_spacing=30):
    """Estimate the place field model.

    Parameters
    ----------
    position : ndarray, shape (n_time,)
    spikes : ndarray, shape (n_time, n_neurons)
    place_bin_centers : ndarray, shape (n_place_bins,)
    penalty : float, optional
    time_bin_size : float, optional

    Returns
    -------
    spiking_likelihood_ratio : function

    """
    min_position, max_position = np.nanmin(position), np.nanmax(position)
    n_steps = (max_position - min_position) // knot_spacing
    position_knots = min_position + np.arange(1, n_steps) * knot_spacing
    FORMULA = ('1 + cr(position, knots=position_knots, constraints="center")')
    training_data = pd.DataFrame(
        dict(position=position[~is_replay])).dropna()
    design_matrix = dmatrix(
        FORMULA, training_data, return_type='dataframe')
    place_field_coefficients = np.stack(
        [fit_glm_model(
            pd.DataFrame(s).loc[design_matrix.index],
            design_matrix, penalty=penalty)
         for s in tqdm(spikes[~is_replay].T, desc='neurons')], axis=1)
    place_design_matrix = create_predict_design_matrix(
        place_bin_centers, design_matrix)
    place_conditional_intensity = get_conditional_intensity(
        place_field_coefficients, place_design_matrix)
    return partial(
        spiking_likelihood_ratio,
        design_matrix=design_matrix,
        place_field_coefficients=place_field_coefficients,
        place_conditional_intensity=place_conditional_intensity,
        time_bin_size=time_bin_size)



"""
"""

from functools import partial

import numpy as np

try:
    from tqdm import tqdm
except ImportError:
    def tqdm(*args, **kwargs):
        if args:
            return args[0]
        return kwargs.get('iterable', None)


def multiunit_likelihood_ratio(multiunit, position, place_bin_centers,
                               place_bins, joint_mark_intensity_functions,
                               ground_process_intensity, time_bin_size=1):
    """The ratio of being in a replay state vs. not a replay state based on
    whether the multiunits correspond to the current position of the animal.

    Parameters
    ----------
    multiunit : ndarray, shape (n_time, n_marks, n_signals)
    position : ndarray, shape (n_time,)
    place_bin_centers : ndarray, shape (n_place_bins,)
    joint_mark_intensity_functions : list of functions
    ground_process_intensity : ndarray, shape (n_place_bins, n_signals)
    time_bin_size : float, optional

    Returns
    -------
    multiunit_likelihood_ratio : ndarray, shape (n_time, n_place_bins)

    """
    position = position
    n_time = position.shape[0]

    replay_log_likelihood = combined_likelihood(
        multiunit, place_bin_centers,
        joint_mark_intensity_functions, ground_process_intensity,
        time_bin_size)
    bin_ind = np.digitize(position, place_bins)
    no_replay_log_likelihood = replay_log_likelihood[
        (np.arange(n_time), bin_ind - 1)][:, np.newaxis]

    return np.exp(replay_log_likelihood - no_replay_log_likelihood)


def combined_likelihood(multiunit, position, joint_mark_intensity_functions,
                        ground_process_intensity, time_bin_size):
    """Combined poisson mark likelihood over all signals.

    Parameters
    ----------
    multiunit : ndarray, shape (n_time, n_marks, n_signals)
    position : ndarray, shape (n_bins, ...)
    joint_mark_intensity_functions : list of functions
    ground_process_intensity : ndarray, shape (n_place_bins, n_signals)
    time_bin_size : float

    Returns
    -------
    combined_likelihood : ndarray, shape (n_time, n_place_bins)

    """
    n_bin = ground_process_intensity.shape[1]
    n_time = multiunit.shape[0]
    log_likelihood = np.zeros((n_time, n_bin))
    multiunit = np.moveaxis(multiunit, -1, 0)

    for signal_marks, jmi, gpi in zip(
            tqdm(multiunit), joint_mark_intensity_functions,
            ground_process_intensity):
        log_likelihood += poisson_mark_log_likelihood(
            jmi(signal_marks, position), gpi,
            time_bin_size)

    return log_likelihood


def poisson_mark_log_likelihood(joint_mark_intensity,
                                ground_process_intensity,
                                time_bin_size=1):
    """Probability of parameters given spiking indicator at a particular
    time and associated marks.

    Parameters
    ----------
    multiunit : array, shape (n_signals, n_marks)
    joint_mark_intensity : function
        Instantaneous probability of observing a spike given mark vector
        from data. The parameters for this function should already be set,
        before it is passed to `poisson_mark_log_likelihood`.
    ground_process_intensity : array, shape (n_signals, n_states,
                                             n_place_bins)
        Probability of observing a spike regardless of multiunit.
    time_bin_size : float, optional

    Returns
    -------
    poisson_mark_log_likelihood : array_like, shape (n_signals, n_states,
                                                     n_time, n_place_bins)

    """
    return np.log(joint_mark_intensity) - (
        ground_process_intensity * time_bin_size)


def joint_mark_intensity(
        multiunit, place_bin_centers, place_occupancy, fitted_model,
        mean_rate):
    """Evaluate the multivariate density function of the marks and place.

    Parameters
    ----------
    multiunit : array, shape (n_time, n_marks)
    place_bin_centers : ndarray, shape (n_place_bins,)
    place_occupancy : ndarray, shape (n_place_bins,)
    fitted_model : a fitted instance of a Class with fit, score_samples, and
                   sample methods.
    mean_rate : float

    Returns
    -------
    joint_mark_intensity : ndarray, shape (n_time, n_place_bins)

    """
    multiunit = np.atleast_2d(multiunit)
    n_place_bins = place_bin_centers.shape[0]
    n_time = multiunit.shape[0]
    is_spike = np.all(~np.isnan(multiunit), axis=1)
    joint_mark_intensity = np.ones((n_time, n_place_bins))

    for ind, position in enumerate(place_bin_centers):
        position = atleast_2d(position * np.ones((n_time,)))
        not_nan = is_spike & ~np.isnan(position.squeeze())
        predict_data = np.concatenate(
            (multiunit[not_nan], position[not_nan]), axis=1)
        joint_mark_intensity[not_nan, ind] = np.exp(
            fitted_model.score_samples(predict_data))

    joint_mark_intensity = mean_rate * joint_mark_intensity / place_occupancy
    joint_mark_intensity[~not_nan] = 1.0
    return joint_mark_intensity


def estimate_place_occupancy(position, place_bin_centers, model, model_kwargs):
    """Probability of being at a position.

    Parameters
    ----------
    position : ndarray, shape (n_time,)
    place_bin_centers : ndarray, shape (n_place_bins,)
    model : Class
    model_kwargs : dict

    Returns
    -------
    place_occupancy : ndarray, shape (n_place_bins,)

    """
    position = atleast_2d(position[~np.isnan(position)])
    return np.exp(model(**model_kwargs).fit(position)
                  .score_samples(place_bin_centers[:, np.newaxis]))


def estimate_ground_process_intensity(
        position, multiunit, place_bin_centers, model, model_kwargs,
        place_occupancy=None):
    """The marginal rate function.

    The probability of observing a spike regardless of mark.

    Parameters
    ----------
    position : ndarray, shape (n_time,)
    multiunit : ndarray, shape (n_time, n_marks)
    place_bin_centers : ndarray, shape (n_place_bins,)
    model : Class
    model_kwargs : dict

    Returns
    -------
    ground_process_intensity : shape (1, n_place_bins)

    """
    if place_occupancy is None:
        place_occupancy = estimate_place_occupancy(
            position, place_bin_centers, model, model_kwargs)

    is_spike = np.all(~np.isnan(multiunit), axis=1)
    not_nan = ~np.isnan(position)
    position = atleast_2d(position)
    place_field = np.exp(model(**model_kwargs)
                         .fit(position[is_spike & not_nan])
                         .score_samples(place_bin_centers[:, np.newaxis]))
    place_occupancy = estimate_place_occupancy(
        position[not_nan], place_bin_centers, model, model_kwargs)
    mean_rate = np.mean(is_spike)
    return np.atleast_2d(mean_rate * place_field / place_occupancy)


def build_joint_mark_intensity(
        position, multiunit, place_bin_centers, model, model_kwargs,
        place_occupancy=None):
    """Make a joint mark intensity function with precalculated quauntities
    set.

    Parameters
    ----------
    position : ndarray, shape (n_time,)
    multiunit : ndarray, shape (n_time, n_marks)
    place_bin_centers : ndarray, shape (n_place_bins,)
    model : Class
    model_kwargs : dict

    Returns
    -------
    joint_mark_intensity : function

    """
    if place_occupancy is None:
        place_occupancy = estimate_place_occupancy(
            position, place_bin_centers, model, model_kwargs)

    multiunit = atleast_2d(multiunit)
    position = atleast_2d(position)

    is_spike = np.all(~np.isnan(multiunit), axis=1)
    not_nan = ~np.isnan(np.squeeze(position))
    mean_rate = np.mean(is_spike, dtype=np.float)

    training_data = np.concatenate(
        (multiunit[is_spike & not_nan], position[is_spike & not_nan]), axis=1)
    fitted_model = model(**model_kwargs).fit(training_data)

    return partial(joint_mark_intensity,
                   place_occupancy=place_occupancy,
                   fitted_model=fitted_model,
                   mean_rate=mean_rate)


def fit_multiunit_likelihood_ratio(position, multiunit, is_replay,
                                   place_bin_centers, place_bins,
                                   model, model_kwargs):
    """Precompute quantities to fit the multiunit likelihood ratio to new data.

    Parameters
    ----------
    position : ndarray, shape (n_time,)
    multiunit : ndarray, shape (n_time, n_marks, n_signals)
    is_replay : bool ndarray, shape (n_time,)
    place_bin_centers : ndarray, shape (n_place_bins,)
    model : Class
    model_kwargs : dict

    Returns
    -------
    multiunit_likelihood_ratio : function

    """
    joint_mark_intensity_functions = []
    ground_process_intensity = []

    position = position[~is_replay]

    place_occupancy = estimate_place_occupancy(
        position, place_bin_centers, model, model_kwargs)

    for m in tqdm(np.moveaxis(multiunit[~is_replay], -1, 0), desc='signals'):
        joint_mark_intensity_functions.append(
            build_joint_mark_intensity(
                position, m, place_bin_centers, model, model_kwargs,
                place_occupancy=place_occupancy))
        ground_process_intensity.append(
            estimate_ground_process_intensity(
                position, m, place_bin_centers, model, model_kwargs,
                place_occupancy=place_occupancy))

    ground_process_intensity = np.concatenate(
        ground_process_intensity, axis=0)

    return partial(
        multiunit_likelihood_ratio,
        place_bin_centers=place_bin_centers,
        place_bins=place_bins,
        joint_mark_intensity_functions=joint_mark_intensity_functions,
        ground_process_intensity=ground_process_intensity,
    )


def atleast_2d(x):
    """Adds a dimension to the last axis if the array is 1D."""
    return np.atleast_2d(x).T if x.ndim < 2 else x



import numpy as np
from functools import wraps


def combined_likelihood(log_likelihood_function):
    @wraps(log_likelihood_function)
    def decorated_function(*args, **kwargs):
        try:
            return np.nansum(log_likelihood_function(*args, **kwargs), axis=-1)
        except ValueError:
            return log_likelihood_function(*args, **kwargs).squeeze()
    return decorated_function


def get_place_bins(position, place_bin_size):
    not_nan_position = position[~np.isnan(position)]
    n_bins = (np.floor(np.ptp(not_nan_position) / place_bin_size) + 2
              ).astype(np.int)
    return np.linspace(
        np.min(not_nan_position), np.max(not_nan_position) + 1E-3, n_bins)


def get_place_bin_centers(bin_edges):
    '''Given the outer-points of bins, find their center
    '''
    return bin_edges[:-1] + np.diff(bin_edges) / 2



import numpy as np
from functools import partial
from sklearn.mixture import GaussianMixture

from spectral_connectivity import Connectivity, Multitaper


def lfp_likelihood_ratio(ripple_band_power, replay_model, no_replay_model):
    """Estimates the likelihood of being in a replay state over time given the
     spectral power of the local field potentials (LFPs).

    Parameters
    ----------
    ripple_band_power : ndarray, shape (n_time, n_signals)
    out_replay_kde : statsmodels.nonparametric.kernel_density.KDEMultivariate
    in_replay_kde : statsmodels.nonparametric.kernel_density.KDEMultivariate

    Returns
    -------
    lfp_likelihood_ratio : ndarray, shape (n_time, 1)

    """
    not_nan = np.all(~np.isnan(ripple_band_power), axis=1)
    n_time = ripple_band_power.shape[0]
    likelihood_ratio = np.ones((n_time, 1))
    no_replay_log_likelihood = no_replay_model.score_samples(
        np.log(ripple_band_power[not_nan]))
    replay_log_likelihood = replay_model.score_samples(
        np.log(ripple_band_power[not_nan]))

    likelihood_ratio[not_nan, 0] = np.exp(
        replay_log_likelihood - no_replay_log_likelihood)
    return likelihood_ratio


def fit_lfp_likelihood_ratio(ripple_band_power, is_replay,
                             model=GaussianMixture,
                             model_kwargs=dict(n_components=3)):
    """Fits the likelihood of being in a replay state over time given the
     spectral power of the local field potentials (LFPs).

    Parameters
    ----------
    ripple_band_power : ndarray, shape (n_time, n_signals)
    is_replay : bool ndarray, shape (n_time,)
    sampling_frequency : float

    Returns
    -------
    likelihood_ratio : function

    """

    not_nan = np.all(~np.isnan(ripple_band_power), axis=1)
    ripple_band_power = np.log(ripple_band_power[not_nan])
    is_replay = is_replay[not_nan]
    replay_model = model(**model_kwargs).fit(
        ripple_band_power[is_replay])
    no_replay_model = model(**model_kwargs).fit(
        ripple_band_power[~is_replay])

    return partial(lfp_likelihood_ratio, replay_model=replay_model,
                   no_replay_model=no_replay_model)


def estimate_ripple_band_power(lfps, sampling_frequency):
    """Estimates the 200 Hz power of each LFP.

    Parameters
    ----------
    lfps : ndarray, shape (n_time, n_signals)
    sampling_frequency : float

    Returns
    -------
    ripple_band_power : ndarray (n_time, n_signals)

    """
    n_time = lfps.shape[0]
    m = Multitaper(lfps, sampling_frequency=sampling_frequency,
                   time_halfbandwidth_product=1,
                   time_window_duration=0.020,
                   time_window_step=1 / sampling_frequency)
    c = Connectivity.from_multitaper(m)
    closest_200Hz_freq_ind = np.argmin(np.abs(c.frequencies - 200))
    power = c.power()[..., closest_200Hz_freq_ind, :].squeeze()
    n_power_time = power.shape[0]
    unobserved = np.full((n_time - n_power_time, *power.shape[1:]), np.nan)
    return np.concatenate((power, unobserved))



import numpy as np
from scipy.stats import norm

RIPPLE_FREQUENCY = 200


def simulate_time(n_samples, sampling_frequency):
    return np.arange(n_samples) / sampling_frequency


def mean_squared(x):
    return (np.abs(x) ** 2.0).mean()


def normalize(y, x=None):
    """normalize power in y to a (standard normal) white noise signal.
    Optionally normalize to power in signal `x`.
    The mean power of a Gaussian with :math:`\\mu=0` and :math:`\\sigma=1` is
    1.
    https://github.com/python-acoustics/python-acoustics/tree/master/acoustics
    """
    x = mean_squared(x) if x is not None else 1.0
    return y * np.sqrt(x / mean_squared(y))


def pink(N, state=None):
    """
    Pink noise.

    :param N: Amount of samples.
    :param state: State of PRNG.
    :type state: :class:`np.random.RandomState`

    Pink noise has equal power in bands that are proportionally wide.
    Power density decreases with 3 dB per octave.
    https://github.com/python-acoustics/python-acoustics/tree/master/acoustics
    """
    state = np.random.RandomState() if state is None else state
    uneven = N % 2
    X = (state.randn(N // 2 + 1 + uneven) +
         1j * state.randn(N // 2 + 1 + uneven))
    S = np.sqrt(np.arange(len(X)) + 1.0)  # +1 to avoid divide by zero
    y = (np.fft.irfft(X / S)).real
    if uneven:
        y = y[:-1]
    return normalize(y)


def white(N, state=None):
    """
    White noise.

    :param N: Amount of samples.
    :param state: State of PRNG.
    :type state: :class:`np.random.RandomState`

    White noise has a constant power density. It's narrowband spectrum is
    therefore flat. The power in white noise will increase by a factor of two
    for each octave band and therefore increases with 3 dB per octave.

    """
    state = np.random.RandomState() if state is None else state
    return state.randn(N)


def brown(N, state=None):
    """
    Brown noise.

    :param N: Amount of samples.
    :param state: State of PRNG.
    :type state: :class:`np.random.RandomState`

    Power decreases with -3 dB per octave.
    Power density decreases with 6 dB per octave.
    """
    state = np.random.RandomState() if state is None else state
    uneven = N % 2
    X = state.randn(N // 2 + 1 + uneven) + 1j * \
        state.randn(N // 2 + 1 + uneven)
    S = np.arange(len(X)) + 1
    y = np.fft.irfft(X / S).real
    if uneven:
        y = y[:-1]
    return normalize(y)


NOISE_FUNCTION = {
    'white': white,
    'pink': pink,
    'brown': brown,
}


def simulate_LFP(time, ripple_times, ripple_amplitude=2,
                 ripple_width=0.025, noise_type='brown', noise_amplitude=1.3):
    '''Simulate a LFP with a ripple at ripple times
    '''
    noise = (noise_amplitude / 2) * NOISE_FUNCTION[noise_type](time.size)
    ripple_signal = np.sin(2 * np.pi * time * RIPPLE_FREQUENCY)
    signal = []

    try:
        iter(ripple_times)
    except TypeError:
        ripple_times = [ripple_times]

    for ripple_time in ripple_times:
        carrier = norm(loc=ripple_time, scale=ripple_width / 6).pdf(time)
        carrier /= carrier.max()
        signal.append((ripple_amplitude / 2) * (ripple_signal * carrier))

    return np.sum(signal, axis=0) + noise



from functools import partial

import numpy as np
import pandas as pd
from patsy import build_design_matrices, dmatrices, NAAction
from statsmodels.api import families
from statsmodels.tsa.tsatools import lagmat
from regularized_glm import penalized_IRLS


def fit_replay_state_transition(speed, is_replay, penalty=1E-5):
    """Estimate the predicted probablity of replay given speed and whether
    it was a replay in the previous time step.

    p(I_t | I_t-1, v_t-1)

    p_I_0, p_I_1 in Long Tao's code

    Parameters
    ----------
    speed : ndarray, shape (n_time,)
    is_replay : boolean ndarray, shape (n_time,)

    Returns
    -------
    probability_replay : ndarray, shape (n_time, 2)

    """
    data = pd.DataFrame({
        'is_replay': is_replay.astype(np.float64),
        'lagged_is_replay': lagmat(
            is_replay, maxlag=1).astype(np.float64).squeeze(),
        'lagged_speed': lagmat(speed, maxlag=1).squeeze()
    }).dropna()
    MODEL_FORMULA = (
        'is_replay ~ 1 + lagged_is_replay + '
        'cr(lagged_speed, knots=[1, 2, 3, 20], constraints="center")')
    response, design_matrix = dmatrices(MODEL_FORMULA, data)
    family = families.Binomial()
    penalty = np.ones((design_matrix.shape[1],)) * penalty
    penalty[0] = 0.0
    fit = penalized_IRLS(design_matrix, response, family=family,
                         penalty=penalty)
    return partial(predict_probability, design_matrix=design_matrix,
                   coefficients=fit.coefficients)


def make_design_matrix(lagged_is_replay, lagged_speed, design_matrix):
    predict_data = {
        'lagged_is_replay': lagged_is_replay * np.ones_like(lagged_speed),
        'lagged_speed': lagged_speed
    }
    return build_design_matrices(
        [design_matrix.design_info], predict_data,
        NA_action=NAAction(NA_types=[]))[0]


def predict_probability(lagged_speed, design_matrix, coefficients):
    """Predict probability of replay state given speed and whether it was a
    replay in the previous time step.

    Parameters
    ----------
    design_matrix : patsy design matrix
    fit : statsmodels fitted model
    speed : ndarray, shape (n_time,)

    Returns
    -------
    replay_probability : ndarray, shape (n_time,)

    """
    no_previous_replay_design_matrix = make_design_matrix(
        0, lagged_speed, design_matrix)

    previous_replay_design_matrix = make_design_matrix(
        1, lagged_speed, design_matrix)

    replay_probability = np.stack(
        (predict(no_previous_replay_design_matrix, coefficients),
         predict(previous_replay_design_matrix, coefficients)),
        axis=1)

    replay_probability[np.isnan(replay_probability)] = 0

    return replay_probability


def predict(design_matrix, coefficients):
    family = families.Binomial()
    return family.link.inverse(design_matrix @ np.squeeze(coefficients))



#!/usr/bin/env python3

from setuptools import find_packages, setup

INSTALL_REQUIRES = ['numpy >= 1.11', 'statsmodels', 'numba', 'matplotlib',
                    'xarray', 'scipy', 'scikit-learn', 'regularized_glm',
                    'spectral_connectivity', 'tqdm']
TESTS_REQUIRE = ['pytest >= 2.7.1']

setup(
    name='replay_identification',
    version='0.1.4.dev0',
    license='MIT',
    description=('Identify replay events using multiple information sources'),
    author='Eric Denovellis',
    author_email='edeno@bu.edu',
    url='https://github.com/Eden-Kramer-Lab/replay_identification',
    packages=find_packages(),
    install_requires=INSTALL_REQUIRES,
    tests_require=TESTS_REQUIRE,
)



import numpy as np
from scipy.stats import multivariate_normal, norm


def simulate_poisson_spikes(rate, sampling_frequency):
    return 1.0 * (np.random.poisson(rate / sampling_frequency) > 0)


def create_place_field(
    place_field_mean, linear_distance, sampling_frequency, is_condition=None,
        place_field_std_deviation=12.5, max_firing_rate=20,
        baseline_firing_rate=0.0):
    if is_condition is None:
        is_condition = np.ones_like(linear_distance, dtype=bool)
    field_firing_rate = norm(
        place_field_mean, place_field_std_deviation).pdf(linear_distance)
    field_firing_rate /= field_firing_rate.max()
    field_firing_rate[~is_condition] = 0
    return baseline_firing_rate + max_firing_rate * field_firing_rate


def generate_marks(spikes, mark_mean, mark_std_deviation, n_marks=4):
    '''Generate a place field with an associated mark'''
    spikes[spikes == 0] = np.nan
    marks = multivariate_normal(
        mean=[mark_mean] * n_marks,
        cov=[mark_std_deviation] * n_marks).rvs(size=(spikes.size,))
    return marks * spikes[:, np.newaxis]


def simulate_multiunit(
    place_field_means, mark_means, linear_distance, sampling_frequency,
        mark_std_deviation=20, n_marks=4, **kwargs):
    '''Simulate a single tetrode assuming each tetrode picks up several
    neurons with different place fields with distinguishing marks.'''
    unit = []
    for place_field_mean, mark_mean in zip(place_field_means, mark_means):
        rate = create_place_field(
            place_field_mean, linear_distance, sampling_frequency, **kwargs)
        spikes = simulate_poisson_spikes(rate, sampling_frequency)
        marks = generate_marks(
            spikes, mark_mean, mark_std_deviation, n_marks=n_marks)
        unit.append(marks)

    return np.nanmean(np.stack(unit, axis=0), axis=0)



"""Calculates the evidence of being in a replay state based on the
current speed and the speed in the previous time step.

"""
from functools import partial

import numpy as np
from statsmodels.api import GLM, families
from statsmodels.tsa.tsatools import lagmat
from patsy import dmatrices


def speed_log_likelihood(response, predicted_response, weights=1., scale=1.):
    '''Gaussian log likelihood by observation

    Parameters
    ----------
    response : ndarray, shape (n_observations,)
    predicted_response : ndarray, shape (n_observations,)
    weights : ndarray or float, shape (n_observations,)
    scale : float

    Returns
    -------
    log_likelihood : ndarray, shape (n_observations,)

    '''
    ll_obs = -weights * (response - predicted_response) ** 2 / scale
    ll_obs += -np.log(scale / weights) - np.log(2 * np.pi)
    ll_obs /= 2
    return ll_obs


def speed_likelihood_ratio(speed, lagged_speed, replay_fit,
                           no_replay_fit, speed_threshold=4.0):
    """Calculates the evidence of being in a replay state based on the
    current speed and the speed in the previous time step.

    Parameters
    ----------
    speed : ndarray, shape (n_time,)
    lagged_speed : ndarray, shape (n_time,)
    replay_speed_std : float
    no_replay_speed_std : float
    speed_threshold : float, optional

    Returns
    -------
    speed_likelihood_ratio : ndarray, shape (n_time, 1)

    """
    no_replay_prediction = no_replay_fit.predict(lagged_speed)
    replay_prediction = replay_fit.predict(lagged_speed)

    replay_log_likelihood = speed_log_likelihood(
        speed, replay_prediction, scale=replay_fit.scale)
    no_replay_log_likelihood = speed_log_likelihood(
        speed, no_replay_prediction, scale=no_replay_fit.scale)
    log_likelihood_ratio = replay_log_likelihood - no_replay_log_likelihood

    likelihood_ratio = np.exp(log_likelihood_ratio)
    likelihood_ratio[np.isposinf(likelihood_ratio)] = 1

    return likelihood_ratio[:, np.newaxis]


def fit_speed_likelihood_ratio(speed, is_replay, speed_threshold=4.0):
    """Fits the standard deviation of the change in speed for the replay and
    non-replay state.

    Parameters
    ----------
    speed : ndarray, shape (n_time,)
    is_replay : ndarray, shape (n_time,)
    speed_threshold : float, optional

    Returns
    -------
    speed_likelihood_ratio : function

    """
    lagged_speed = lagmat(speed, 1)
    replay_fit = fit_speed_model(speed[is_replay], lagged_speed[is_replay])
    no_replay_fit = fit_speed_model(
        speed[~is_replay], lagged_speed[~is_replay])
    return partial(speed_likelihood_ratio,
                   replay_fit=replay_fit,
                   no_replay_fit=no_replay_fit,
                   speed_threshold=speed_threshold)


def fit_speed_model(speed, lagged_speed):
    FORMULA = 'speed ~ lagged_speed - 1'
    response, design_matrix = dmatrices(
        FORMULA, dict(speed=speed, lagged_speed=lagged_speed))
    family = families.Gaussian(link=families.links.log)
    return GLM(response, design_matrix, family=family).fit()
